from semanticnetwork import TangledSemanticNetwork
from nltk.stem.wordnet import WordNetLemmatizer as WNL
import copy

DECLARATIVE = 0
IMPERATIVE = 1
QUESTION = 2

class ParseError(Exception):
    def __init__(self, value):
        self.value = value

class NounPhrase:
    def __init__(self):                             
        self.pronoun = None
        self.proper_noun = None
        self.noun = None
        self.article = None
        self.adjs = []
        self.np = None
        self.pp = None

    def get_subject(self):
        if self.noun:
            return self.noun
        elif self.pronoun:
            return self.pronoun
        elif self.proper_noun:
            return self.proper_noun

class VerbPhrase:
    def __init__(self):
        self.verb = None
        self.np = None
        self.adjs = []
        self.pp = None
        self.adv = None
        self.vp = None
                        
class PrepositionPhrase:
    def __init__(self):
        self.prep = None
        self.np = None

class Sentence:
    def __init__(self, unparsed_sentence, kb):
        self.kb_full = kb
        self.kb = kb.language
        self.NP_FIRSTSET = set(["noun", "pronoun", "proper noun", "article", "adjective"]) 
        self.INT_PRONOUNS_FIRSTSET = set(["what", "who", "where", "when", "why", "how"])
        self.INT_VERBS_FIRSTSET = set(["was", "do", "does", "can", "should", "would", "could", "is", "have", "are"])
        self.vp = None
        self.np = None
        self.int_verb = None
        self.sentence = None
        self.is_object = None
        self.parseS(self.listify(unparsed_sentence))

    def listify(self, sentence):
        return listify(self.kb, sentence)
    
    def get_sentence_type(self):
        if self.int_verb:
            return QUESTION
        elif self.np:
            return DECLARATIVE
        else:
            return IMPERATIVE


    def parseS(self, sentence):
        """ S : NP VP """
        fst_pos_set = self.get_pos_set(sentence[0])
        if sentence[0] in self.INT_PRONOUNS_FIRSTSET:
            self.int_pronoun = sentence.pop(0)
            if sentence[0] in self.INT_VERBS_FIRSTSET:
                self.int_verb = sentence.pop(0)
                self.np = self.parseNP(sentence)
                if len(sentence) > 0:
                    self.vp = self.parseVP(sentence)
            elif "verb" in self.get_pos_set(sentence[0]):
                self.vp = self.parseVP(sentence)
        elif self.has_common_elt(fst_pos_set, self.NP_FIRSTSET):
            self.np = self.parseNP(sentence)
            self.vp = self.parseVP(sentence)
        elif sentence[0] in self.INT_VERBS_FIRSTSET:
            self.int_verb = sentence.pop(0)
            self.np = self.parseNP(sentence)
            if self.int_verb != "is":
                self.vp = self.parseVP(sentence)
            else:
                self.is_object = self.parseNP(sentence)
        elif "verb" in fst_pos_set:
            self.vp = self.parseVP(sentence)
        else:
            raise ParseError("Bad sentence")
        if len(sentence) > 0:
            if "connective" in self.get_pos_set(sentence[0]):
                self.connective = sentence.pop(0)
                self.sentence = Sentence(' '.join(sentence), self.kb_full)
            else:
                raise ParseError("Sentence not completely consumed")
        return self
            

    def parseNP(self, sentence):
        """ NP : pronoun | proper noun | noun | ArtP | NP_Helper """ 
        np = NounPhrase()
        if len(sentence) > 0:
            fst_pos_set = self.get_pos_set(sentence[0])
            if "noun" in fst_pos_set:
                np.noun = sentence.pop(0)
            elif "pronoun" in fst_pos_set:
                np.pronoun = sentence.pop(0)
            elif "proper noun" in fst_pos_set:
                np.proper_noun = sentence.pop(0)
            elif "article" in fst_pos_set:
                np.article = sentence.pop(0)
                np = self.parseArtP(sentence, np)        
            elif "adjective" in fst_pos_set:
                np.adjs = self.parseAdjs(sentence)
            else:
                raise ParseError("Bad sentence")
            return self.parseNP_Helper(sentence, np)
        raise ParseError("Sentenced prematurely consumed")

    def parseVP(self, sentence):
        if len(sentence) > 0:
            if "verb" in self.get_pos_set(sentence[0]):
                vp = VerbPhrase()
                vp.verb = sentence.pop(0)
                if vp.verb == "is":
                    vp = self.parseVP_Helper(sentence, vp, True)
                else:
                    vp = self.parseVP_Helper(sentence, vp)
                return vp
            else:
                raise ParseError("Bad sentence")
        raise ParseError("Sentenced prematurely consumed")

    def parsePP(self, sentence):
        if len(sentence) > 0:
            if "preposition" in self.get_pos_set(sentence[0]):
                pp = PrepositionPhrase()
                pp.prep = sentence.pop(0)
                pp.np = self.parseNP(sentence)
                return pp
            else:
                raise ParseError("Bad sentence")    
        raise ParseError("Sentenced prematurely consumed")

    def parseAdjs(self, sentence):
        """ Ajds : Ajd Ajds | e """
        if len(sentence) > 0:
            fst_pos_set = self.get_pos_set(sentence[0])
            if "adjective" in fst_pos_set:
                adj = sentence.pop(0)
                adjs = self.parseAdjs(sentence)
                adjs.append(adj)
                return adjs
            return []
        raise ParseError("Sentenced prematurely consumed")

# Helper function to deal with ambiguity in NP
# Returns a tuple (Adjs, noun); Adjs will be None
# if input matches first production 
    def parseArtP(self, sentence, np):
        """ ArtP : noun | Adjs noun """
        if len(sentence) > 0:
            fst_pos_set = self.get_pos_set(sentence[0])
            if "noun" in fst_pos_set:
                np.noun = sentence.pop(0)
            elif "adjective" in fst_pos_set:
                np.adjs = self.parseAdjs(sentence)
                np.noun = sentence.pop(0)
            else:
                raise ParseError(sentence[0])
            return np
        raise ParseError("Sentenced prematurely consumed")
        
# NP_Helper added to deal with left recursion in
# NP productions (e.g. the production NP PP)
    def parseNP_Helper(self, sentence, np, err_set=None):
        if err_set == None:
            err_set = set([])
        if len(sentence) > 0:     
            fst_pos_set = self.get_pos_set(sentence[0]).difference(err_set)
            sentence_snapshot = list(sentence)
            np_snapshot = copy.copy(np)
            try: 
                if "noun" in fst_pos_set:
                    np.np = self.parseNP(sentence)
                    np.pp = self.parsePP(sentence)
                elif "preposition" in fst_pos_set:
                    np.pp = self.parsePP(sentence)
            except ParseError as e:
                sentence = sentence_snapshot            
                np = np_snapshot
                err_set.add("noun")
                return self.parseNP_Helper(sentence, np, err_set)
        return np

    def parseVP_Helper(self, sentence, vp, adj_can_follow=None):
        if len(sentence) > 0:
            fst_pos_set = self.get_pos_set(sentence[0])
            if "adjective" in fst_pos_set and adj_can_follow:
                vp.adjs = [sentence.pop(0)]
            elif self.has_common_elt(fst_pos_set, self.NP_FIRSTSET):
                vp.np = self.parseNP(sentence)
            elif "preposition" in fst_pos_set:
                vp.pp = self.parsePP(sentence)
            elif "adverb" in fst_pos_set:
                vp.adv = sentence.pop(0)
            elif "connective" in fst_pos_set:
                return vp
            elif "verb" in fst_pos_set and vp.verb in self.INT_VERBS_FIRSTSET:
                vp.vp = self.parseVP(sentence)
            else:
                raise ParseError("Bad sentence")
            return self.parseVP_Helper(sentence, vp)
        return vp

    def get_pos_set(self, word):
        paths = self.kb.paths(word, 'partofspeech')
        pos_set = set([])
        for path in paths:
            pos_set.add(path[1])
        return pos_set
        #return self.kb.succeeds(word)

    def has_common_elt(self, set1, set2):
        for x in set1:
            if x in set2:
                return True
        return False

def listify(kb, sentence):
    lm = WNL()
    sentence = sentence.split()
    
    sentence = map(lambda word: word.rstrip(",.?!"), sentence)
    sentence = map(lambda word: word.lower(), sentence)
    
    def lemmatize(word):
        root_word = lm.lemmatize(word)
        
        if kb.paths(root_word, "partofspeech"):
            return root_word
        else:
            root_word = lm.lemmatize(word, 'v') 
            if kb.paths(root_word, "partofspeech"):
                return root_word
            else:
                return word
        
    sentence = map(lambda word: lemmatize(word), sentence)
    return sentence
