from functools import partial
from util import list_prepend, list_concat

class NetworkNode(object):
    """
    A node in a graph.
    """
    
    def __init__(self, value):
        self.descendents = set()
        self.predecessors = set()
        self.val = value
    
    def add_link(self, other):
        if other not in self.descendents:
            self.descendents.add(other)
            other.predecessors.add(self)

    def break_link(self, other):
        if other in self.descendents:
            self.descendents.discard(other)
            other.predecessors.discard(self)

    def has_link(self, other):
        return other in self.descendents

    def __contains__(self, other):
        return self.has_link(other)

    def __hash__(self):
        return hash(self.val)

    def __eq__(self, other):
        if type(other) == NetworkNode:
            return (self.val == other.val and
                    self.descendents == other.descendents)
        else:
            return self.val == other

    def __str__(self):
        return "NetworkNode(%s)" % self.val

    def __repr__(self):
        return self.__str__()

    def __ge__(self, other):
        return self.val < other.val

    def __lt__(self, other):
        return self.val < other.val
    
    def __iter__(self):
        return iter(self.descendents)


class TangledSemanticNetwork(object):
    """
    A semantic network that expresses arbitrary relationships
    between nodes. A node is a subtype of another node if it is a
    child of that node.
    
    This class supports various syntactic manipulations:

    >>>'foo' in tsn
    true if foo is a node in this network, false otherwise
    
    >>>tsn['cat'] = 'pet'
    analogous to link('cat', 'pet')

    >>>tsn['cat']
    returns a list of all nodes linked to from cat, like 'pet'
    """

    def __init__(self, root=None):
        self.root = None
        self.lookup = {}
        if root:
            self.root = NetworkNode(root)
            self.lookup[root] = self.root
    
    def __contains__(self, val):
        if type(val) == NetworkNode:
            val = val.val
        return val in self.lookup

    def link(self, fv, to):
        """
        Link fv to to in the network, creating nodes for both of them
        if necessary.
        """
        tn = self.lookup[to] if to in self else NetworkNode(to)
        if not fv in self:
            self.add(fv)
        self.lookup[fv].add_link(tn)
        self.lookup[to] = tn

    def add(self, v):
        """
        Add v to the network as an isolated node.
        """
        self.lookup[v] = NetworkNode(v)

    def succeeds(self, a, b=None):
        """
        Returns true if b exists somewhere on the graph above a
        (for example, pet is above cat)

        If b is not given, returns the set of nodes immediately above a.
        """
        return self._on_path_from(a, b, lambda n: n.descendents)
                                  

    def precedes(self, a, b=None):
        """
        Returns true if b is below a -- for example, cat is below pet.

        If b is not given, returns the set of nodes immediately below a.
        """
        return self._on_path_from(a, b, lambda n: n.predecessors)

    def _on_path_from(self, a, b, extend):
        if not b: 
            return extend(self.lookup[a])

        search = [self.lookup[a].val]
        while len(search) > 0:
            s = search.pop(0)
            if s == b:
                return True
            search.extend(extend(self.lookup[s]))
        return False

    def __setitem__(self, fv, t):
        self.link(fv, t)

    def __getitem__(self, v):
        if not v in self:
            self.add(v)
        return map(lambda n: n.val,
            list(self.lookup[v].descendents))

    def unlink(self, p, c):
        """
        Break a link between node p and node c.
        """
        pn = self.lookup[p]
        cn = self.lookup[c]
        pn.break_link(cn)
        if len(cn.predecessors) == 0:
            del self.lookup[c]

    def paths(self, f, t):
        """
        Return a list of ordered paths from node f to node t.
        """
        if not f in self or not t in self:
            # non-existant node, return nothing
            return []
        f = self.lookup[f]
        t = self.lookup[t]
        # if we're here, return a one-hop list
        if f == t:
            return [[t.val]]
        # otherwise, add ourselves to every non-empty list
        ret = []
        for fp in f:
            ret.extend(map(partial(list_prepend, f.val), 
                           filter(lambda l: len(l) > 0, 
                                  self.paths(fp.val, t.val))))
        return ret

    def contains_link(self, f, t):
        return f in self and t in self and \
            t in self.lookup[f]
        

# this makes debugging easier
tsn = TangledSemanticNetwork

